#include "common.c"

int short_flowlabel(int sock, struct sockaddr_in6 *dst, __u32 label)
{
    int on = 1;
    struct in6_flowlabel_req freq;

    memset(&freq, 0, sizeof(freq));
    freq.flr_label = htonl(label);
    freq.flr_action = IPV6_FL_A_GET;
    freq.flr_flags = IPV6_FL_F_CREATE;
    freq.flr_share = IPV6_FL_S_EXCL;
    freq.flr_linger = 10;
#ifdef EXPTIMER
    freq.flr_expires = EXPTIMER;
#else
    freq.flr_expires = 15;
#endif
    memcpy(&freq.flr_dst, &dst->sin6_addr, 16);
    if (setsockopt(sock, SOL_IPV6, IPV6_FLOWLABEL_MGR, &freq, sizeof(freq)) == -1) {
	perror("can't lease flowlabel");
	return -1;
    }
    dst->sin6_flowinfo = freq.flr_label;
    if (setsockopt(sock, SOL_IPV6, IPV6_FLOWINFO_SEND, &on, sizeof(on)) == -1) {
	perror("can't send flowinfo");
	freq.flr_action = IPV6_FL_A_PUT;
	setsockopt(sock, SOL_IPV6, IPV6_FLOWLABEL_MGR, &freq, sizeof(freq));
	dst->sin6_flowinfo = 0;
	return -1;
    }
    return 0;
}

int main(int ac,char **av)
{
    __u32 flowlabel = 0;
    int sock, len, err_num, itmp;
    socklen_t sin6_len = 0;
    uint16_t port = 0;
    struct in6_addr addr;
    struct sockaddr_in6 sin6, d_sin6;
    char *target, v6addr[128];

    sock = len = err_num = itmp = 0;
    if (ac != 4) {
	printf("usage: %s <flowlabel> <dst_host> <dst_port>\n",av[0]);
	return 0;
    }

    sscanf(av[1],"%x",&flowlabel);
    if (flowlabel == 0) {
	printf("flowlabel is null.\n");
	return -1;
    }
    target = av[2];
    sscanf(av[3],"%d",&itmp);
    if (itmp == 0) {
	printf("dst_port is null.\n");
	return -2;
    } else {
	port = (uint16_t)itmp;
    }

    if (getip6name(&addr, target, "dst_addr") < 0) return -3;
    sin6_len = sizeof(sin6);
    memset(&sin6, 0, sin6_len);
    sin6.sin6_family = AF_INET6;
    memcpy(&sin6.sin6_addr, &addr, 16);
    sin6.sin6_port = htons(port);
    sock = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
    if (sock < 0) {
	printf("cannot open INET6 socket.\n");
	return -4;
    }
    if (set_rxflow(sock) < 0) {
	printf("can't get flowlabel\n");
    }
    if (short_flowlabel(sock, &sin6, flowlabel) < 0) {
	printf("cannot use flowlabel.\n");
	close(sock);
	return -5;
    }
    if (connect(sock, (struct sockaddr *)&sin6, sin6_len) == -1) {
	perror("connect");
	err_num = -6;
	goto err;
    }

    len = send(sock, target, strlen(target), 0);
    if (len < 0) {
	perror("send");
	err_num = -7;
	goto err;
    }

    len = 0; memset(v6addr, 0, 128);
    memcpy(&d_sin6, &sin6, sin6_len);
    d_sin6.sin6_flowinfo = 0;
#if 0
//    len = recvfrom(sock, v6addr, 128, 0, (struct sockaddr*)&sin6, &sin6_len);
    len = recvlabel(sock, v6addr, 128, 0, (struct sockaddr*)&sin6, &sin6_len);
    if (len < 0) {
	perror("recvfrom");
	err_num = -8;
	goto err;
    }
#else
    len = recv(sock, v6addr, 128, 0);
//    if (getpeername(sock, (struct sockaddr*)&d_sin6, &sin6_len) == -1) {
    if (getlabel(sock, &d_sin6, &sin6_len) < 0) {
	perror("getpeername");
	return -21;
    }
#endif
    printf("dst_host flow has 0x%x.\n",ntohl(d_sin6.sin6_flowinfo));
    printf("message: %s", v6addr);
    sleep(2);
 err:
    release_flowlabel(sock, &sin6, flowlabel);
    close(sock);
    return err_num;
}
